#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "freetype.h"

/*
 * Einfaches Programm zur Ausgabe eines gedrehten TT-Strings 
 * Schriftverzeichnis, Schrift und Gre sind hart eingestellt
 * Der String kann als Parameter angegeben werden.
 * Fehlt er, wird "Test" angezeigt.
 * Ist ein zweiter Parameter angegeben, wird dieser als
 * Drehwinkel (im mathematischen Sinn) interpretiert. 
 * Fehlt er, wird der der Text um 12 gedreht
 * Das Fenster wird geschlossen, wenn der Benutzer
 * eine Taste drckt, whrend es den Fokus hat
 */

#define FONTDIR "/usr/X11R6/lib/X11/fonts/ttfonts/"
#define FONT FONTDIR "arial.ttf"
#define SIZE 120

#define FLOOR(x) ( (x) & -64)
#define CEIL(x)  ( ((x) + 63) & -64)

static void show_bitmap(char *bitmap, int lines, int cols);

main(int argc, unsigned char *argv[]) {
  TT_Error error;
  TT_Face face;
  TT_Face_Properties properties;
  TT_Instance instance;
  TT_Engine engine;
  TT_Glyph glyph;
  TT_UShort platform, encoding;
  TT_CharMap cmap;
  TT_Raster_Map Bit;
  TT_UShort left, right;
  TT_UShort nPairs;
  TT_Matrix italic;
  TT_Outline outline;

  unsigned char *test="Test";
  unsigned char *ptr;
  float angle = 12.0;
  int width, height;
  int minx, miny, maxx, maxy;
  TT_Pos xoffset, yoffset;
  register int i;
  
  if (argc > 1) 
    test = argv[1];
  if (argc > 2)
    angle = atof(argv[2]);
  error = TT_Init_FreeType(&engine);
  if (error) {
    printf("Keine TT-Maschine erzeugt\n");
    exit(1);
  }
  error=TT_Open_Face( engine, FONT, &face);
  if (error) {
    printf("Font nicht initialisiert\n");
    exit(1);
  }
 /* Charmap 1 ist Windows/Unicode - normalerweise
     mte man hier suchen */
  TT_Get_CharMap(face, 1, &cmap);
  error=TT_New_Instance( face, &instance);
  if (error) {
    printf("Instanz nicht initialisiert\n");
    exit(1);
  }
  /* Bildschirm */
  error=TT_Set_Instance_Resolutions(instance, 72, 72);
  if (error) {
    printf("Fehler beim Einstellen der Auflsung\n");
    exit(1);
  }
  error=TT_Set_Instance_CharSize(instance, SIZE*64);
  if (error) {
    printf("Fehler beim Einstellen der Gre\n");
    exit(1);
  }
  error=TT_New_Glyph(face, &glyph);
  if (error) {
    printf("Fehler beim Einstellen der Gre\n");
    exit(1);
  }
  width = height = 0;
  minx = miny = maxx = maxy = 0;
  angle = angle * 3.14159265/512;
  italic.xx = (TT_Fixed)(cos(angle) *  (1<<16));
  italic.xy = (TT_Fixed) (sin( angle ) * (1<<16));
  italic.yx = - italic.xy;
  italic.yy = italic.xx; 

  xoffset = yoffset = 0;

  for (ptr=test; *ptr; ptr++) {
    TT_UShort mapped;
    TT_Pos vx, vy;
    TT_Glyph_Metrics metric;
    TT_BBox bbox;

    mapped = TT_Char_Index(cmap, (TT_UShort)*ptr);
    error  = TT_Load_Glyph(instance, glyph, mapped,
			   TTLOAD_DEFAULT);
    TT_Get_Glyph_Metrics(glyph, &metric);
    TT_Get_Glyph_Outline(glyph, &outline);
    TT_Transform_Outline(&outline, &italic);
    TT_Translate_Outline(&outline, xoffset, yoffset);
    TT_Get_Outline_BBox(&outline, &bbox);
    
    bbox.xMin = FLOOR(bbox.xMin);
    bbox.yMin = FLOOR(bbox.yMin);
    bbox.xMax = CEIL(bbox.xMax);
    bbox.yMax = CEIL(bbox.yMax);

    vx = metric.advance;
    vy = 0;
    TT_Transform_Vector(&vx, &vy, &italic);

    xoffset += vx;
    yoffset += vy;
    
    width = xoffset;

    if (miny > bbox.yMin)
      miny = bbox.yMin;
    if (maxy < bbox.yMax)
      maxy = bbox.yMax;
    if (minx > bbox.xMin)
      minx = bbox.xMin;

  } /* for ... */
  height = maxy - miny;

  Bit.rows = height / 64;
  Bit.width = width / 64;
  Bit.flow   = TT_Flow_Down;
  Bit.cols = (Bit.width + 7) >> 3;
  Bit.size = (long)Bit.cols * Bit.rows;
  Bit.bitmap = calloc(sizeof(char), (int)Bit.size);
  xoffset = abs(minx);
  yoffset = abs(miny);

  for (ptr=test; *ptr; ptr++) {
    TT_UShort mapped;
    TT_Pos vx, vy;
    TT_Glyph_Metrics metric;
    TT_BBox bbox;

    mapped = TT_Char_Index(cmap, (TT_UShort)*ptr);
    error  = TT_Load_Glyph(instance, glyph, mapped,
			   TTLOAD_DEFAULT);
    TT_Get_Glyph_Metrics(glyph, &metric);
    TT_Get_Glyph_Outline(glyph, &outline);
    TT_Transform_Outline(&outline, &italic);
    TT_Translate_Outline(&outline, xoffset, yoffset);
    
    TT_Get_Outline_Bitmap(engine,&outline, &Bit);
    vx = metric.advance;
    vy = 0;
    TT_Transform_Vector(&vx, &vy, &italic);

    xoffset += vx;
    yoffset += vy;
  }
  show_bitmap(Bit.bitmap, Bit.rows, Bit.width);
}


static void show_bitmap( char *bitmap, int lines, int cols) {
  Display *display;
  XImage *image;
  char *ptr;
  Window win;
  GC gc;
  XSetWindowAttributes  xswa;
  int screen_num, depth;
  int win_width, win_height;
  int xoff, yoff;
  Visual*   visual;
  Colormap  colormap;
  char      *windowmap, *target;
  register int i;
  
  display = XOpenDisplay(NULL);
  
  screen_num = DefaultScreen  ( display );
  colormap   = DefaultColormap( display, screen_num );
  depth      = DefaultDepth   ( display, screen_num );
  visual     = DefaultVisual  ( display, screen_num );

  xswa.border_pixel     = BlackPixel( display, screen_num );
  xswa.background_pixel = WhitePixel( display, screen_num );
  
  xswa.event_mask = KeyPressMask | ExposureMask;
  
  win_width  = cols * 1.5;
  win_height = lines * 1.5;
  win = XCreateWindow( display,
		       RootWindow( display, screen_num ),
		       0,
		       0,
		       win_width,
		       win_height,
		       10,
		       depth,
		       InputOutput, 
		       visual,
		       CWBackPixel | CWBorderPixel | CWEventMask,
		       &xswa );
  
  XMapWindow( display, win );
  gc = XCreateGC( display, RootWindow( display, screen_num ),
		       0L, NULL );

  XSetForeground( display, gc, BlackPixel( display, screen_num ) );
  XSetBackground( display, gc, WhitePixel( display, screen_num ) );

  image = XCreateImage( display, 
			visual,
			1, 
			XYBitmap,
			0, 
			bitmap,
			cols,
			lines,
			8, 
			0 );
  image->byte_order       = MSBFirst;
  image->bitmap_bit_order = MSBFirst;
  while (1) {
    XEvent event;
    XNextEvent(display, &event);
    switch (event.type) {
      case Expose:
	xoff = (win_width - cols)/2;
	yoff = (win_height - lines) / 2;
	XPutImage(display, win, gc, image, 0,0,
		  xoff, yoff,  cols, lines);
	break;
      case KeyPress:
	exit(0);
    }
  }/* while 1 */
}
