#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include "freetype.h"

/*
 * Einfaches Programm zur Ausgabe von "Guten Tag" auf
 * Griechisch und Russisch in Arial
 * Schriftverzeichnis, Schrift, Gre und Text sind hart eingestellt
 * Das Fenster wird geschlossen, wenn der Benutzer
 * eine Taste drckt, whrend es den Fokus hat
 */

#define FONTDIR "/usr/X11R6/lib/X11/fonts/ttfonts/"
#define FONT FONTDIR "arial.ttf"
#define SIZE 48

static void show_bitmap(char *bitmap, int lines, int cols);

main(int argc, unsigned char *argv[]) {
  TT_Error error;
  TT_Face face;
  TT_Face_Properties properties;
  TT_Instance instance;
  TT_Engine engine;
  TT_Glyph glyph;
  TT_Glyph_Metrics metric;
  TT_UShort platform, encoding;
  TT_CharMap cmap;
  TT_Raster_Map Bit;

  int width, height;
  int xoffset, yoffset;
  int minx, miny, maxx, maxy;
  register int i;
  static TT_UShort test[] = {0x39a, 0x3b1,0x3bb,0x3b7,0x3bc,0x3b5,0x3c1,0x3b1,
			     0x20, (int)'-',0x20,
			     0x414, 0x43e, 0x431, 0x440, 0x42b,0x20,
			     0x414,0x435,0x43d, 0x44c, (int)'!',0};
  TT_UShort *ptr;

  error = TT_Init_FreeType(&engine);
  if (error) {
    printf("Keine TT-Maschine erzeugt\n");
    exit(1);
  }
 
  error=TT_Open_Face( engine, FONT, &face);
  if (error) {
    printf("Font nicht initialisiert\n");
    exit(1);
  }
  error = TT_Get_Face_Properties(face, &properties);
  for (i = 0; i < properties.num_CharMaps; i++) {
    TT_Get_CharMap_ID(face, (TT_UShort)i, &platform, &encoding);
    printf("Charmap %d: Platform %d, Encoding %d\n", i, 
	   (int)platform, (int)encoding);
  }
  TT_Get_CharMap(face, 1, &cmap);
  error=TT_New_Instance( face, &instance);
  if (error) {
    printf("Instanz nicht initialisiert\n");
    exit(1);
  }
  /* Bildschirm */
  error=TT_Set_Instance_Resolutions(instance, 72, 72);
  if (error) {
    printf("Fehler beim Einstellen der Auflsung\n");
    exit(1);
  }
  error=TT_Set_Instance_CharSize(instance, SIZE*64);
  if (error) {
    printf("Fehler beim Einstellen der Gre\n");
    exit(1);
  }
  error=TT_New_Glyph(face, &glyph);
  if (error) {
    printf("Fehler beim Einstellen der Gre\n");
    exit(1);
  }
  width = height = 0;
  minx = miny = maxx = maxy = 0;
  for (ptr = test; *ptr; ptr++) {
    TT_UShort mapped;
    mapped = TT_Char_Index(cmap, (TT_UShort)*ptr);
    error = TT_Load_Glyph(instance, glyph, mapped,
			  TTLOAD_DEFAULT);
    TT_Get_Glyph_Metrics(glyph, &metric);
    width += metric.advance;
    if (miny > metric.bbox.yMin)
      miny = metric.bbox.yMin;
    if (maxy < metric.bbox.yMax)
      maxy = metric.bbox.yMax;
    

  }
  miny = (miny & -64);
  maxy = ( (maxy + 63) & -64);

  height = maxy - miny;

  Bit.rows = (height + 32) / 64;
  Bit.width = (width +32) / 64;
  Bit.flow   = TT_Flow_Down;
  Bit.cols = (Bit.width + 7) >> 3;
  Bit.size = (long)Bit.cols * Bit.rows;
  Bit.bitmap = calloc(sizeof(char), (int)Bit.size);
  xoffset = 0;
  yoffset = abs(miny);
  for (ptr = test; *ptr; ptr++) {
    TT_UShort mapped;
    mapped = TT_Char_Index(cmap, (TT_UShort)*ptr);
    error = TT_Load_Glyph(instance, glyph, mapped,
			  TTLOAD_DEFAULT);
    TT_Get_Glyph_Metrics(glyph, &metric);
    TT_Get_Glyph_Bitmap(glyph, &Bit, xoffset, yoffset);
    xoffset += metric.advance;
  }
  show_bitmap(Bit.bitmap, Bit.rows, Bit.width);
}


static void show_bitmap( char *bitmap, int lines, int cols) {
  Display *display;
  XImage *image;
  char *ptr;
  Window win;
  GC gc;
  XSetWindowAttributes  xswa;
  int screen_num, depth;
  int win_width, win_height;
  Visual*   visual;
  Colormap  colormap;
  char      *windowmap, *target;
  register int i;
  
  display = XOpenDisplay(NULL);
  
  screen_num = DefaultScreen  ( display );
  colormap   = DefaultColormap( display, screen_num );
  depth      = DefaultDepth   ( display, screen_num );
  visual     = DefaultVisual  ( display, screen_num );

  xswa.border_pixel     = BlackPixel( display, screen_num );
  xswa.background_pixel = WhitePixel( display, screen_num );
  
  xswa.event_mask = KeyPressMask | ExposureMask;
  
  win_width =  cols * 1.5;
  win_height = lines * 1.5;
  win = XCreateWindow( display,
		       RootWindow( display, screen_num ),
		       0,
		       0,
		       win_width,
		       win_height,
		       10,
		       depth,
		       InputOutput, 
		       visual,
		       CWBackPixel | CWBorderPixel | CWEventMask,
		       &xswa );
  
  XMapWindow( display, win );
  
  gc = XCreateGC( display, RootWindow( display, screen_num ),
		       0L, NULL );
  XSetForeground( display, gc, BlackPixel( display, screen_num ) );
  XSetBackground( display, gc, WhitePixel( display, screen_num ) );
  image = XCreateImage( display, 
			visual,
			1, 
			XYBitmap,
			0, 
			bitmap,
			cols,
			lines,
			8, 
			0 );
 
  image->byte_order       = MSBFirst;
  image->bitmap_bit_order = MSBFirst;
 
  while (1) {
    XEvent event;
    XNextEvent(display, &event);
    switch (event.type) {
      case Expose:
	XPutImage(display, win, gc, image, 0,0,
		  (win_width - cols)/2, (win_height - lines) /2,
		  cols, lines);
	break;
      case KeyPress:
	exit(0);
    }
  }/* while 1 */
}
